// -*- Mode: C++; -*-
//                            	Package   : omniORB
// SocketCollection.h		Created on: 23 Jul 2003
//				Author    : Serguei Kolos
//
//    Copyright (C) 2001 AT&T Laboratories Cambridge
//
//    This file is part of the omniORB library
//
//    The omniORB library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Library General Public
//    License as published by the Free Software Foundation; either
//    version 2 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Library General Public License for more details.
//
//    You should have received a copy of the GNU Library General Public
//    License along with this library; if not, write to the Free
//    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//    02111-1307, USA
//
//
// Description:
//	*** PROPRIETORY INTERFACE ***
//

/*
*/

#ifndef __SOCKETCOLLECTION_H__
#define __SOCKETCOLLECTION_H__

////////////////////////////////////////////////////////////////////////
//  Platform feature selection

#define SOCKNAME_SIZE_T OMNI_SOCKNAME_SIZE_T

#define USE_NONBLOCKING_CONNECT

#if defined(__linux__)
#   define USE_POLL
#endif

#if defined(__sunos__)
#   define USE_POLL
#endif

#if defined(__hpux__)
#   if __OSVERSION__ >= 11
#       define USE_POLL
#   endif
#   define USE_FAKE_INTERRUPTABLE_RECV
#endif

#if defined(__freebsd__)
#  define USE_POLL
#endif

#if defined(__WIN32__)
#   define USE_FAKE_INTERRUPTABLE_RECV
#endif

#if defined(__irix__)
#   define USE_POLL
#endif

////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
//             win32 API
//
#if defined(__WIN32__)

#  include <sys/types.h>
#  include <libcWrapper.h>

#  define RC_INADDR_NONE     INADDR_NONE
#  define RC_INVALID_SOCKET  INVALID_SOCKET
#  define RC_SOCKET_ERROR    SOCKET_ERROR
#  define INETSOCKET         PF_INET
#  define CLOSESOCKET(sock)  closesocket(sock)
#  define SHUTDOWNSOCKET(sock) ::shutdown(sock,2)
#  define ERRNO              ::WSAGetLastError()
#  define EINPROGRESS        WSAEWOULDBLOCK
#  define RC_EINTR           WSAEINTR
#  define RC_EBADF           WSAENOTSOCK
#  define NEED_SOCKET_SHUTDOWN_FLAG 1

OMNI_NAMESPACE_BEGIN(omni)

typedef SOCKET SocketHandle_t;

OMNI_NAMESPACE_END(omni)

#else

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
//             unix(ish)
//
#  if defined(__vxWorks__)
#    include <sockLib.h>
#    include <hostLib.h>
#    include <ioLib.h>
#    include <netinet/tcp.h>
#  else
#    include <sys/time.h>
#  endif
#  include <sys/socket.h>
#  include <netinet/in.h>
#  include <netinet/tcp.h>
#  include <arpa/inet.h>
#  include <unistd.h>
#  include <sys/types.h>
#  include <errno.h>
#  include <libcWrapper.h>

#  if defined(USE_POLL)
#    include <poll.h>
#  endif

#  if !defined(__VMS)
#    include <fcntl.h>
#  endif

#  if defined (__uw7__)
#    ifdef shutdown
#      undef shutdown
#    endif
#  endif

#  if defined(__VMS) && defined(USE_tcpSocketVaxRoutines)
#    include "tcpSocketVaxRoutines.h"
#    undef accept
#    undef recv
#    undef send
#    define accept(a,b,c) tcpSocketVaxAccept(a,b,c)
#    define recv(a,b,c,d) tcpSocketVaxRecv(a,b,c,d)
#    define send(a,b,c,d) tcpSocketVaxSend(a,b,c,d)
#  endif

#  ifdef __rtems__
extern "C" int select (int,fd_set*,fd_set*,fd_set*,struct timeval *);
#  endif

#  define RC_INADDR_NONE     ((CORBA::ULong)-1)
#  define RC_INVALID_SOCKET  (-1)
#  define RC_SOCKET_ERROR    (-1)
#  define INETSOCKET         AF_INET
#  define CLOSESOCKET(sock)  close(sock)

#  if defined(__sunos__) && defined(__sparc__) && __OSVERSION__ >= 5
#    define SHUTDOWNSOCKET(sock)  ::shutdown(sock,2)
#  elif defined(__osf1__) && defined(__alpha__)
#    define SHUTDOWNSOCKET(sock)  ::shutdown(sock,2)
#  else
     // XXX none of the above, calling shutdown() may not have the
     // desired effect.
#    define SHUTDOWNSOCKET(sock)  ::shutdown(sock,2)
#  endif

#  define ERRNO              errno
#  define RC_EINTR           EINTR
#  define RC_EBADF           EBADF

OMNI_NAMESPACE_BEGIN(omni)

typedef int    SocketHandle_t;

OMNI_NAMESPACE_END(omni)

#endif

#if defined(NEED_GETHOSTNAME_PROTOTYPE)
extern "C" int gethostname(char *name, int namelen);
#endif

OMNI_NAMESPACE_BEGIN(omni)

/////////////////////////////////////////////////////////////////////////
// This bit is set in the mask field of the indfd 
// structure to indicate that this fd has not to be 
// taken into account while processing the poll result
#define FDCLRD		0x01
// This bit is set in the events field of the pollfd 
// structure to indicate that this fd has to be 
// moved to the end of the pollfd array
#define FDMVD		0x02
// This bit is set in the events field of the pollfd 
// structure to indicate that this fd has data to be read
#define FDDIB		0x04

// This class is a replacement for the two fd sets used
// by the select-based SocketCollection implementation
// It holds an array of pollfd structures, which consits
// from two parts:
// 1. elements [0..length_now-1] will be used 
//    for the next poll invocation
// 2. elements [length_now..length_full-1] will be passed 
//    to poll only if the Reschedule method has been called
class SocketHandleSet
{
public:
  
  SocketHandleSet();
  ~SocketHandleSet();
	
  // Adds socket to the array of pollfd structures 
  // If now == 1, this socket is set to the first part of the 
  //   array and will be used for the next poll invocation
  // If now == 0 the socket will be stored in the last part and will
  //   be passed to poll only if the Reschedule method is called
  // If data_in_buffer == 1 the FDDTNBFFR bit will be set in the
  //   events field of the corresponding pollfd structure
  void Add( SocketHandle_t fd, CORBA::Boolean now, CORBA::Boolean data_in_buffer );
  
  // Removes socket from the array of pollfd structures 
  void Remove( SocketHandle_t fd );

  // Moves socket from the current position to the end of array of pollfd structures 
  void Move( SocketHandle_t fd );

  // Checks if socket exist in the array of pollfd structures 
  // If data_in_buffer == 1, function returns 1 only if socket has the FDDTNBFFR
  //   bit set in the events field of the corresponding pollfd structure
  // If data_in_buffer == 0, function returns 1 if socket is in 
  //   the array of pollfd structures
  CORBA::Boolean isSet( SocketHandle_t fd, CORBA::Boolean data_in_buffer );
	  
  // Checks if socket has-to-be removed 
  inline CORBA::Boolean htbRemoved( SocketHandle_t fd ) {
    return ( pd_index[fd].mask & FDCLRD );
  }
	  
  // Checks if socket has-to-be moved 
  inline CORBA::Boolean htbMoved( SocketHandle_t fd ) {
    return ( pd_index[fd].mask & FDMVD );
  }
	  
  // Checks if socket has data in buffer bit set
  inline CORBA::Boolean hasDIB( SocketHandle_t fd ) {
    return ( pd_index[fd].mask & FDDIB );
  }
	  
  // Must be called before entering the poll
  inline pollfd * enterPoll( unsigned int & length ) {
    length = pd_length_now;
    pd_used_by_poll_n = pd_length_now;
    return pd_data;
  }

  // Must be called after exiting the poll
  inline void exitPoll( ) {
    pd_used_by_poll_n = 0;
  }
  
  // Schedule all the sockets for the next poll invocation
  inline void Reschedule( ) {
    pd_length_now = pd_length_full;
  }
  
  // Returns 1 if there are sockets in the array, which were 
  // requested to be moved or removed, or if there are 
  // unprocessed sockets marked as having data in buffer
  inline CORBA::Boolean needsProcessing( ) {
    return (    pd_to_be_removed_n
	     || pd_to_be_moved_n
	     || pd_dib_n );
  }
	
  struct indfd {
    short index;
    char  mask;
  };					// takes more memory, but allows mark fds
  
private:
  struct 
  pollfd *		pd_data;
  indfd *		pd_index;	// takes more memory, but allows to find
  					// sockets int the data array very efficiently
  unsigned short	pd_length_full;
  unsigned short	pd_length_now;
  unsigned short	pd_used_by_poll_n;
  unsigned short	pd_to_be_removed_n;
  unsigned short	pd_to_be_moved_n;
  unsigned short	pd_dib_n;
};


class SocketCollection;

extern void SocketSetTimeOut(unsigned long abs_sec,
			     unsigned long abs_nsec,struct timeval& t);

extern int SocketSetnonblocking(SocketHandle_t sock);

extern int SocketSetblocking(SocketHandle_t sock);

extern int SocketSetCloseOnExec(SocketHandle_t sock);

class SocketLink {

public:
  SocketLink(SocketHandle_t s)
    : pd_socket(s),
#ifdef NEED_SOCKET_SHUTDOWN_FLAG
      pd_shutdown(0),
#endif
      pd_next(0) {}

  ~SocketLink() {}

  friend class SocketCollection;

protected:
  SocketHandle_t pd_socket;

#ifdef NEED_SOCKET_SHUTDOWN_FLAG
  // select() on Windows does not return an error after the socket has
  // shutdown, so we have to store an extra flag here.
  CORBA::Boolean pd_shutdown;
#endif

private:
  SocketLink*    pd_next;
};

class SocketCollection {
public:

  SocketCollection();

protected:
  virtual ~SocketCollection();

  virtual CORBA::Boolean notifyReadable(SocketHandle_t) = 0;
  // Callback used by Select(). This method is called while holding
  // pd_fdset_lock.

public:
  void setSelectable(SocketHandle_t sock, CORBA::Boolean now,
		     CORBA::Boolean data_in_buffer,
		     CORBA::Boolean hold_lock=0);
  // Indicates that this socket should be watched by a poll()
  // so that any new data arriving on the connection will be noted.
  // If now == 1, immediately make this socket part of the poll
  // set.
  // If data_in_buffer == 1, treat this socket as if there are
  // data available from the connection already.
  // If hold_lock == 1, pd_fdset_lock is already held.

  void clearSelectable(SocketHandle_t);
  // Indicates that this connection need not be watched any more.

  CORBA::Boolean Select();
  // Returns TRUE(1) if the Select() has successfully done a scan.
  // otherwise returns false(0) to indicate that an error has been
  // detected and this function should not be called again.
  //
  // For each of the sockets that has been marked watchable and indeed
  // has become readable, call notifyReadable() with the socket no.
  // as the argument.

  CORBA::Boolean Peek(SocketHandle_t sock);
  // Do nothing and returns immediately if the socket has not been
  // set to be watched by a previous setSelectable().
  // Otherwise, monitor the socket's status for a short time.
  // Returns TRUE(1) if the socket becomes readable.
  // otherwise returns FALSE(0).

  void incrRefCount();
  void decrRefCount();

  void addSocket(SocketLink* conn);
  // Add this socket to the collection. <conn> is associated with the
  // socket and should be added to the table hashed by the socket number.
  // Increments this collection's refcount.

  SocketLink* removeSocket(SocketHandle_t sock);
  // Remove the socket from this collection. Return the socket which has
  // been removed. Return 0 if the socket is not found.
  // Decrements this collection's refcount if a socket is removed.

  SocketLink* findSocket(SocketHandle_t sock,
			 CORBA::Boolean hold_lock=0);
  // Returns the connection that is associated with this socket.
  // Return 0 if this socket cannot be found in the hash table.
  // if hold_lock == 1, the caller has already got the lock on pd_fdset_lock.
  // (use purely by member functions.)

  static unsigned long scan_interval_sec;
  static unsigned long scan_interval_nsec;

  static CORBA::ULong  hashsize;

private:
  SocketHandleSet      pd_fdset;
  omni_tracedmutex     pd_fdset_lock;
  omni_tracedcondition pd_poll_cond; // timedwait on if nothing to select
  unsigned long        pd_abs_sec;
  unsigned long        pd_abs_nsec;
  int                  pd_pipe_read;
  int                  pd_pipe_write;
  CORBA::Boolean       pd_pipe_full;
  int                  pd_refcount;

protected:
  SocketLink**         pd_hash_table;

};

OMNI_NAMESPACE_END(omni)

#endif // __SOCKETCOLLECTION_H__
